///////////////////////////////////////////////////////////////////////////////////
//No part of this file can be copied or released without the consent of 
//Avalanche Technology
//										
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
//										
//	Avalanche Technology Inc., Proprietary and Confidential	   *
//										
// 	Release:  2.1    Date 7/21/2021  	
//							
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
//  PART DESCRIPTION:
//
//  Technology: 40nm pMTJ STT-MRAM
//  Part:       AS3016204
//
//  Description: 16 Megabit HIGH PERFORMANCE SERIAL PERSISTANT SRAM MEMORY
//
////////////////////////////////////////////////////////////////////////////////////
//  FILE CONTENTS : MODEL OF SPSRAM 
//
////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////
// MODULE DECLARATION                                                             //
////////////////////////////////////////////////////////////////////////////////////

`timescale 100ps / 10ps

module spsram_as3016204
	( 
		IO3, 
		WPn_IO2, 
		SO_IO1,
		SI_IO0,
		CSn, 
		CLK
	);
	
	inout IO3; 
	inout SI_IO0; 
	inout SO_IO1;
	inout WPn_IO2;
	input CSn;
	input CLK;
	//input [7:0] mode;

	reg hold_dq3_reg; 
	reg si_dq0_reg; 
	reg so_dq1_reg;
	reg wp_dq2_reg;
	
	reg dataout_neg;

	reg [23:0] prot_addr_min ; 
	reg [23:0] prot_addr_max ;
	
    	//Status Register 
    	reg [7:0] SR     = 8'h00;	
	
    	//Config Registers 
    	reg [31:0] CR1     = 8'h00;	
    	reg [31:0] CR2     = 8'h00;
    	reg [31:0] CR3     = 8'h00;
    	reg [31:0] CR4     = 8'h00;

    	// MEMORY DEFINITION
	parameter mem_dim = 24'h1FFFFF;
	parameter addr_mask = 24'h1FFFFF;
	reg [7:0] mem_data [(mem_dim-1):0] ;
	
	
	reg [23:0] loc = 0;

	// Instruction type
	parameter NOP             = 8'h00;    
	parameter WREN            = 8'h06;
	parameter WRDI            = 8'h04;
	parameter WRSR            = 8'h01;	
	parameter RDSR            = 8'h05;
	parameter WRCR            = 8'h87;	
	parameter RDCR            = 8'h46;
	parameter READ            = 8'h03;
	parameter FAST_READ       = 8'h0b;
	parameter WRITE           = 8'h02;


	// States
	parameter CMD             = 8'h00;    
	parameter ADDR1           = 8'h01;  
	parameter ADDR2           = 8'h02;  
	parameter ADDR3           = 8'h03;  
	parameter MODE_BYTE       = 8'h04;  
	parameter LATENCY      	  = 8'h05;  
	parameter DATAIN      	  = 8'h06;
	parameter DATAOUT      	  = 8'h07;

	//SPI DPI and QPI modes
	parameter SPIEN 	= 8'hFF;
	parameter DPIEN 	= 8'h37;
	parameter QPIEN 	= 8'h38;
	
	// Registers
    	reg [7:0] datain_reg;
    	reg [7:0] dataout_reg;
    	reg [7:0] dataout_reg_0;
    	reg [7:0] dataout_reg_1;
    	reg [7:0] dataout_reg_2;
    	reg [7:0] dataout_reg_3;

      	reg [2:0] datain_bit_count;
	reg [2:0] dataout_bit_count;
      	reg [1:0] datain_bit_count_dpi;
	reg [1:0] dataout_bit_count_dpi;
      	reg datain_bit_count_qpi;
	reg dataout_bit_count_qpi;
	reg [23:0] MEM_ADD;

	// Interface Type
	reg [1:0] Interface_type;
	reg [7:0] latency_count;
	reg PORn;
	reg [7:0] state;
	reg [7:0] mode;
        reg [7:0] next;
        reg [7:0] cmd;
	reg [23:0] addr_count;
	reg addr_load_first;

	wire [7:0] cmd_comb_spi;
	wire [7:0] cmd_comb_dpi;
	wire [7:0] cmd_comb_qpi;

	wire cmd_phase ;
        wire addr1_phase ;
        wire addr2_phase ;
        wire addr3_phase ;
	wire mode_phase ;
	wire latency_phase ;
	wire datain_phase ;
	wire dataout_phase ;
	wire datain_latchen;
	wire addrin_latchen;
	wire mem_read;
	wire sr_read;
	wire cfg_read;
	wire read_strobe;
	wire read_strobe_dpi;
	wire read_strobe_qpi;
	wire sck_dly_pad;
	wire first_mem_access;
	wire first_mem_access_dpi;
	wire first_mem_access_qpi;
	wire next_state;
	wire spi_mode;
	wire dpi_mode;
	wire qpi_mode;
	wire curr_mode;
	////////////////////////////////////////////////////////////
	// Timing Violation checks
	///////////////////////////////////////////////////////////
	//$setup, $hold, $width violations
	reg tCSS_notifier;
	reg tCSH_notifier;
	reg tCH_notifier;
	reg tCL_notifier;
	reg tCS_notifier;
	reg tSU_notifier;
	reg tHD_notifier;

	specify 
		specparam tPOR_time  = 1; 
		specparam tCSS       = 50; 
		specparam tCSH       = 40; 
		specparam tCH        = 46; 
		specparam tCL        = 46; 
		specparam tCS        = 1300; 
		specparam tSU        = 20;
		specparam tHD        = 20;
		specparam tCO        = 70;

  		$setup (negedge CSn, posedge CLK, tCSS,  tCSS_notifier);
  		$hold  (negedge CLK, posedge CSn, tCSH,  tCSH_notifier);

  		$width (posedge CLK, tCH, 0, tCH_notifier); 
 		$width (negedge CLK, tCL, 0, tCL_notifier);
		$width (posedge CSn, tCS, 0, tCS_notifier);

  		$setup (SI_IO0, posedge CLK, tSU,  tSU_notifier);
		$hold  (posedge CLK, SI_IO0, tHD,  tHD_notifier);

	endspecify 
	
// Power on reset
    	initial
    	begin
		for (loc = 0; loc < mem_dim ; loc = loc+1)
		begin
			mem_data[loc] = 8'h00;
		end
        	PORn = 1'b0;
        	#tPOR_time PORn = 1'b1;
                mode = SPIEN;
    	end
	
	// Initial Reset
	always @(posedge PORn)
	begin
		Interface_type <= 2'b00;
		MEM_ADD <= 24'h0;
		dataout_reg <= 8'h00;
		dataout_reg_0 <= 8'h00;
		dataout_reg_1 <= 8'h00;
		dataout_reg_2 <= 8'h00;
		dataout_reg_3 <= 8'h00;
		hold_dq3_reg <= 0; 
		si_dq0_reg <= 0; 
		so_dq1_reg <= 0;
		wp_dq2_reg <= 0;
	end

	assign next_state = (datain_bit_count == 3'b000 & CSn == 1'b0) ? 1'b1 : 1'b0;
	assign next_state = (datain_bit_count_dpi == 2'b00 & CSn == 1'b0) ? 1'b1 : 1'b0;
	assign next_state = (datain_bit_count_qpi == 1'b0 & CSn == 1'b0) ? 1'b1 : 1'b0;
	assign cmd_phase = (state == CMD) ? 1'b1 : 1'b0;
        assign addr1_phase = (state == ADDR1) ? 1'b1 : 1'b0;
        assign addr2_phase = (state == ADDR2) ? 1'b1 : 1'b0;
        assign addr3_phase = (state == ADDR3) ? 1'b1 : 1'b0;
	assign mode_phase = (state == MODE_BYTE) ? 1'b1 : 1'b0;
	assign latency_phase = (state == LATENCY) ? 1'b1 : 1'b0;
	assign datain_phase = (state == DATAIN ) ? 1'b1 : 1'b0;
	assign dataout_phase = (state == DATAOUT) ? 1'b1 : 1'b0;
	assign datain_latchen = (datain_bit_count == 3'b000 && datain_phase == 1) ? 1'b1 : 1'b0;
	assign datain_latchen = (datain_bit_count_dpi == 2'b00 && datain_phase == 1) ? 1'b1 : 1'b0;
	assign datain_latchen = (datain_bit_count_qpi == 1'b0 && datain_phase == 1) ? 1'b1 : 1'b0;
	assign #(tCO) SO_IO1 = (dataout_neg == 1'b1 && CSn != 1 && state!=ADDR3) ? so_dq1_reg : 1'bz;
	assign #(tCO) SI_IO0 = (dataout_neg == 1'b1 && CSn != 1 && state!=ADDR3) ? si_dq0_reg : 1'bz;
	assign #(tCO) WPn_IO2 = (dataout_neg == 1'b1 && CSn != 1 && state!=ADDR3) ? wp_dq2_reg : 1'bz;
	assign #(tCO) IO3 = (dataout_neg == 1'b1 && CSn != 1 && state !=ADDR3) ? hold_dq3_reg : 1'bz;
	assign read_strobe = ((addr3_phase == 1'b1 && datain_bit_count == 8'h7) || (dataout_phase == 1'b1 && dataout_bit_count == 4'h7)) ? 1'b1 : 1'b0;
	assign read_strobe_dpi = ((addr3_phase == 1'b1 && datain_bit_count_dpi == 8'h3) || (dataout_phase == 1'b1 && dataout_bit_count_dpi == 4'h3)) ? 1'b1 : 1'b0;
	assign read_strobe_qpi = ((addr3_phase == 1'b1 && datain_bit_count_qpi == 8'h1) || (dataout_phase == 1'b1 && dataout_bit_count_qpi == 4'h1)) ? 1'b1 : 1'b0;
	assign mem_read = (cmd == READ || cmd == FAST_READ) ? read_strobe : 1'b0;
	assign mem_read_dpi = (cmd == READ || cmd == FAST_READ) ? read_strobe_dpi : 1'b0;
	assign mem_read_qpi = (cmd == READ || cmd == FAST_READ) ? read_strobe_qpi : 1'b0;
	assign first_mem_access = (addr3_phase == 1'b1 && datain_bit_count == 8'h0) ? 1'b1 : 1'b0;
	assign first_mem_access_dpi = (addr3_phase == 1'b1 && datain_bit_count_dpi == 8'h0) ? 1'b1 : 1'b0;
	assign first_mem_access_qpi = (addr3_phase == 1'b1 && datain_bit_count_qpi == 8'h0) ? 1'b1 : 1'b0;

	assign spi_mode = (mode == SPIEN) ? 1'b1 : 1'b0;
	assign dpi_mode = (mode == DPIEN) ? 1'b1 : 1'b0;
	assign qpi_mode = (mode == QPIEN) ? 1'b1 : 1'b0;
	assign curr_mode = (spi_mode == 1'b1 || dpi_mode == 1'b1 || qpi_mode == 1'b1) ? 1'b1 : 1'b0;


	always @(negedge CLK or posedge CSn)
	begin
		if (CSn == 1'b1)
			dataout_neg <= 1'b0;
		else if (next == DATAOUT)
			dataout_neg <= 1'b1;
		else
			dataout_neg <= 1'b0;
	end



	// SPI Interface
	always @(posedge CLK or posedge CSn)
	begin
		if(CSn) begin
			datain_reg <= 8'h00;
			datain_bit_count <= 3'h0;
			datain_bit_count_dpi <= 2'h0;
			datain_bit_count_qpi <= 1'h0;
		end
		else if (mode == SPIEN)begin
			datain_reg <= {datain_reg[6:0],SI_IO0};
			datain_bit_count <= datain_bit_count + 7;
		end
		else if (mode == DPIEN)begin
			datain_reg <= {datain_reg[5:0],SO_IO1,SI_IO0};
			datain_bit_count_dpi <= datain_bit_count_dpi + 3;
		end
		else if (mode == QPIEN)begin 
			datain_reg <= {datain_reg[3:0],IO3,WPn_IO2,SO_IO1,SI_IO0};
			datain_bit_count_qpi <= datain_bit_count_qpi + 1;
		end	
	end
	
	always @(posedge CLK or posedge CSn or posedge curr_mode)
	begin
		if(CSn)
			state <= CMD;
		else if((next_state && latency_phase ==0) || (latency_count == (CR2[3:0]-1) && latency_phase == 1))
			state <= next;
		else
			state <= state;
			
	end
	
	always @(posedge CLK or posedge CSn)
	begin
             	if(CSn) begin
                   	cmd <= NOP;
		end
		else if(state == CMD && datain_bit_count == 3'h1 && mode == SPIEN)begin
                       	cmd <= {datain_reg[6:0],SI_IO0};
             	end
		else if(state == CMD && datain_bit_count_dpi == 2'h1 && mode == DPIEN)begin
			cmd <= {datain_reg[5:0],SO_IO1,SI_IO0};
		end
		else if(state == CMD && datain_bit_count_qpi == 1'h1 && mode == QPIEN)begin
			cmd <= {datain_reg[3:0],IO3,WPn_IO2,SO_IO1,SI_IO0};
		end
		else begin
			cmd <= cmd;
		end
	end

	always @ (posedge CLK or posedge CSn or posedge curr_mode)
	begin
		if(latency_phase == 1'b1)
			latency_count <= latency_count +1;
		else
			latency_count <= 0;
	end

	always @(cmd or state or CSn or curr_mode)
	begin

		if(state == CMD) begin
			if(mode == SPIEN && (cmd == DPIEN||cmd ==QPIEN))begin
				mode = cmd;
				if(cmd == DPIEN )
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b1;
					$display("[MODEL] IO mode set to DPI");
				end
				else if (cmd == QPIEN )
				begin
                                	CR2[6] = 1'b1;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to QPI");
				end
				else 
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to SPI");
				end
			end
			else if (mode == DPIEN && (cmd == SPIEN||cmd ==QPIEN))begin
				mode = cmd;
				if(cmd == DPIEN )
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b1;
					$display("[MODEL] IO mode set to DPI");
				end
				else if (cmd == QPIEN )
				begin
                                	CR2[6] = 1'b1;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to QPI");
				end
				else 
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to SPI");
				end
			end
			else if (mode == QPIEN && (cmd == DPIEN||cmd ==SPIEN))begin
				mode = cmd;
				if(cmd == DPIEN )
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b1;
					$display("[MODEL] IO mode set to DPI");
				end
				else if (cmd == QPIEN )
				begin
                                	CR2[6] = 1'b1;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to QPI");
				end
				else 
				begin
                                	CR2[6] = 1'b0;
					CR2[4] = 1'b0;
					$display("[MODEL] IO mode set to SPI");
				end
			end
			case (cmd)
				NOP : next = CMD;
				WRDI : next = CMD;
				WRSR : next = DATAIN;	
				RDSR : next = DATAOUT;	
				WRCR : next = DATAIN;	
				RDCR : next = DATAOUT;
				READ : next = (mode == SPIEN) ?ADDR1 : CMD;	
				FAST_READ : next = ADDR1;	
				WRITE :	next = ADDR1;	
				default : next = CMD;
			endcase
		end
		else if (state == ADDR1) begin
			case (cmd)
				READ : next = ADDR2;	
				FAST_READ : next = ADDR2;	
				WRITE :	next = ADDR2;	
				default : next = CMD;
			endcase		
		end			
		else if (state == ADDR2) begin
			case (cmd)
				READ : next = ADDR3;	
				FAST_READ : next = ADDR3;	
				WRITE :	next = ADDR3;	
				default : next = CMD;	
			endcase	
		end			
		else if (state == ADDR3) begin
			case (cmd)
				READ : begin
					if (CR2[3:0] !=0) begin 
						next = LATENCY;	
					end
					else begin
						next = DATAOUT;
					end
				end
				FAST_READ : next = MODE_BYTE;	
				WRITE :	next = DATAIN;	
				default : next = CMD;	
			endcase	
		end		
		else if (state == MODE_BYTE) begin
			case (cmd)
				FAST_READ :begin
					if (CR2[3:0] !=0) begin 
						next = LATENCY;	
					end
					else begin
						next = DATAOUT;
					end
				end	
				default : next = CMD;	
			endcase	
		end
		else if(state == LATENCY) begin
			case (cmd)
				READ : next = DATAOUT;	
				FAST_READ : next = DATAOUT;	
				default : next = CMD;	
			endcase	
		end
		else if(state == DATAIN) begin
				next = DATAIN;
		end
		else if(state == DATAOUT) begin
			next = DATAOUT;
		end
		else begin
			next = CMD;
		end
	end

	always @(posedge CLK)
	begin
		if (mode == SPIEN)
		begin
		if(addr1_phase == 1 && datain_bit_count == 0) begin
			MEM_ADD <= {datain_reg[7:0], 16'h0000};
		end 
		else if(addr2_phase == 1 && datain_bit_count == 0) begin
			MEM_ADD <= {MEM_ADD[23:16],datain_reg[7:0], 8'h00};
		end 
		else if(addr3_phase == 1 && datain_bit_count == 6) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[5:0],2'b0};
		end		
		else if(addr3_phase == 1 && datain_bit_count == 0) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[7:0]};
		end
		end
	
		else if (mode == DPIEN)
		begin
		if(addr1_phase == 1 && datain_bit_count_dpi == 0) begin
			MEM_ADD <= {datain_reg[7:0], 16'h0000};
		end 
		else if(addr2_phase == 1 && datain_bit_count_dpi == 0) begin
			MEM_ADD <= {MEM_ADD[23:16],datain_reg[7:0], 8'h00};
		end 
		else if(addr3_phase == 1 && datain_bit_count_dpi == 2) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[5:0],2'b0};
		end		
		else if(addr3_phase == 1 && datain_bit_count_dpi == 0) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[7:0]};
		end
		end
		
		else if (mode == QPIEN)
		begin
		if(addr1_phase == 1 && datain_bit_count_qpi == 0) begin
			MEM_ADD <= {datain_reg[7:0], 16'h0000};
		end 
		else if(addr2_phase == 1 && datain_bit_count_qpi == 0) begin
			MEM_ADD <= {MEM_ADD[23:16],datain_reg[7:0], 8'h00};
		end 
		else if(addr3_phase == 1 && datain_bit_count_qpi == 1) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[5:0],2'b0};
		end		
		else if(addr3_phase == 1 && datain_bit_count_qpi == 0) begin
			MEM_ADD <= {MEM_ADD[23:8],datain_reg[7:0]};
		end
		end

	end

	always @(posedge CLK or posedge CSn)
	begin
		if(CSn) begin
			addr_count <= 0;
 			addr_load_first <= 0;
		end

		else if (mode == SPIEN)
		begin
			if(datain_phase == 1 && datain_bit_count == 3'b000 ) begin
			addr_count <= addr_count + 1;
			end
			else if (dataout_phase == 1 && dataout_bit_count == 3'b111) begin
			addr_count <= addr_count + 1;
			end
		end

 		else if (mode == DPIEN)
		begin
			if(datain_phase == 1 && datain_bit_count_dpi == 2'b00 ) begin
			addr_count <= addr_count + 1;
			end
			else if (dataout_phase == 1 && dataout_bit_count_dpi == 2'b11) begin
			addr_count <= addr_count + 1;
			end
		end

		else if (mode == QPIEN)
		begin
			if(datain_phase == 1 && datain_bit_count_qpi == 1'b0) begin
			addr_count <= addr_count + 1;
			end
			else if (dataout_phase == 1 && dataout_bit_count_qpi == 1'b1) begin
			addr_count <= addr_count + 1;
			end
		end

	end

	always @(negedge CLK or posedge CSn or posedge curr_mode)
	begin
		if(datain_latchen == 1'b1 && CSn == 0)
		begin
			if(cmd == WRSR)	begin
				if( SR[1] == 1'b1 && CR1[2] == 1'b0)
				begin
					SR <= datain_reg & 8'hfc;
					$display("[MODEL] Status Write :- Data %x ",datain_reg);
				end
				else if (CR1[2] == 1'b1)
				begin
					$display("[MODEL] No Status Write :- status register locked");
				end

				else if ( SR[1] == 1'b0)
				begin
					$display("[MODEL] No Status Write :- WREN bit not set ");
				end
			end
			else if (cmd == WRCR)	begin
				if( SR[1] == 1'b1)
				begin
					case(addr_count)
						24'h000000 : CR1 <= datain_reg;
						24'h000001 : CR2 <= CR2 & 8'h50 |datain_reg & 8'haf;
						24'h000002 : CR3 <= datain_reg;
						24'h000003 : CR4 <= datain_reg;
					endcase
					$display("[MODEL] Config Register Write :- CR1 %x CR2 %x CR3 %x CR4 %x",CR1, CR2, CR3, CR4);
				end
				else
				begin
					$display("[MODEL] No Config register Write :- WREN bit not set");
				end
			end
			else if (cmd == WRITE)	begin
				if( SR[1] == 1'b1)
				begin
					//mem_data[MEM_ADD + addr_count] <= datain_reg;
					//$display("[MODEL] Memory Write :- Address %x Data %x Data IN %x ",(MEM_ADD + addr_count),mem_data[MEM_ADD + addr_count],datain_reg);
					if(((MEM_ADD + addr_count) > prot_addr_min) && ((MEM_ADD + addr_count) < prot_addr_max))begin
           				$display("[MODEL]Memory Write: Address is in protect range -Write Address %x protected addr min %x protected addr max %x",(MEM_ADD+addr_count),prot_addr_min,prot_addr_max); 
					end    
					else begin
					mem_data[MEM_ADD + addr_count] <= datain_reg;
					$display("[MODEL] Memory Write :- Address %x Data %x Data IN %x ",(MEM_ADD + addr_count),mem_data[MEM_ADD + addr_count],datain_reg);            
					end

				end
				else
				begin
					$display("[MODEL] No Memory Write :- WREN bit not set");
				end
			end
                        
		end

		if( cmd == WREN && cmd_phase == 1 && curr_mode == 1) begin
			SR <= SR | 8'h02;
		end
		else if( cmd == WRDI && cmd_phase == 1 && curr_mode == 1) begin
			SR <= SR & 8'hfc;
		end
		else if((cmd == WRITE ||  cmd == WRSR || cmd == WRCR) && CSn == 1 && curr_mode == 1)
		begin
			SR <= SR & 8'hfc;
		end

	end

 	always @(SR)
	begin

            		case({SR[5],SR[4:2]})
	 			4'b0000 : begin prot_addr_min = 24'h200000; prot_addr_max = 24'h200000; end// PROTECT_RANGE = NO_PROT
         			4'b0001 : begin prot_addr_min = 24'h1F8000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_64
         			4'b0010 : begin prot_addr_min = 24'h1F0000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_32 
        			4'b0011 : begin prot_addr_min = 24'h1E0000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_16
         			4'b0100 : begin prot_addr_min = 24'h1C0000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_8
         			4'b0101 : begin prot_addr_min = 24'h180000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_4
         			4'b0110 : begin prot_addr_min = 24'h1F0000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = UPPER_1_2
         			4'b0111 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = ALL_PROT
         			4'b1000 : begin prot_addr_min = 24'h200000; prot_addr_max = 24'h200000; end// PROTECT_RANGE = NO_PROT
         			4'b1001 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h007FFF; end// PROTECT_RANGE = LOWER_1_64
         			4'b1010 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h00FFFF; end// PROTECT_RANGE = LOWER_1_32
         			4'b1011 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h01FFFF; end// PROTECT_RANGE = LOWER_1_16
         			4'b1100 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h03FFFF; end// PROTECT_RANGE = LOWER_1_8
         			4'b1101 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h07FFFF; end// PROTECT_RANGE = LOWER_1_4
	 			4'b1110 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h0FFFFF; end// PROTECT_RANGE = LOWER_1_2
	 			4'b1111 : begin prot_addr_min = 24'h000000; prot_addr_max = 24'h1FFFFF; end// PROTECT_RANGE = ALL_PROT
      			endcase
/*			if(((MEM_ADD + addr_count) > prot_addr_min) && ((MEM_ADD + addr_count) < prot_addr_max))begin
            			//mem_data[MEM_ADD + addr_count] <= datain_reg;
           			//$display("[INFO][MODEL]Memory Write: Address is in protect range -Write Address %x protected addr min %x protected addr max %x",(MEM_ADD+addr_count),prot_addr_min,prot_addr_max); 
			end    
			else begin
            
			end
			else begin

			end
*/
	end


	always @(negedge CLK or posedge CSn or posedge curr_mode)
	begin
		if(CSn) begin
			dataout_reg <= 0;
		end
		else begin
			case(cmd)
				RDSR : begin
					dataout_reg <= SR;

				end
				RDCR : begin
					if (addr_count == 24'h000000) begin
						dataout_reg <= CR1;
					end
					else if (addr_count == 24'h000001) begin
						dataout_reg <= CR2;
					end
					else if (addr_count == 24'h000002) begin
						dataout_reg <= CR3;
					end
					else if (addr_count == 24'h000003) begin
						dataout_reg <= CR4;
					end
				end
				READ : begin
						if ((mem_read || mem_read_dpi || mem_read_qpi) && addr3_phase == 1) begin
							dataout_reg_0 <= mem_data[(MEM_ADD  + addr_count) & addr_mask];
							dataout_reg_1 <= mem_data[(MEM_ADD +1 + addr_count) & addr_mask];
							dataout_reg_2 <= mem_data[(MEM_ADD +2 + addr_count) & addr_mask];
							dataout_reg_3 <= mem_data[(MEM_ADD +3 + addr_count) & addr_mask];
						end
						else if ((first_mem_access || first_mem_access_dpi || first_mem_access_qpi) == 1'b1) begin
							case(datain_reg[1:0])
								2'b00: dataout_reg <= dataout_reg_0;
								2'b01: dataout_reg <= dataout_reg_1;
								2'b10: dataout_reg <= dataout_reg_2;
								2'b11: dataout_reg <= dataout_reg_3;
							endcase 
						end
						else if ((mem_read || mem_read_dpi || mem_read_qpi) && addr3_phase == 0) begin
							dataout_reg <= mem_data[(MEM_ADD + addr_count) & addr_mask];
						end

					end
	
				FAST_READ : begin
						if((mem_read || mem_read_dpi || mem_read_qpi) && addr3_phase == 1) begin
							dataout_reg_0 <= mem_data[(MEM_ADD  + addr_count) & addr_mask];
							dataout_reg_1 <= mem_data[(MEM_ADD +1 + addr_count) & addr_mask];
							dataout_reg_2 <= mem_data[(MEM_ADD +2 + addr_count) & addr_mask];
							dataout_reg_3 <= mem_data[(MEM_ADD +3 + addr_count) & addr_mask];
						end
						else if ((first_mem_access || first_mem_access_dpi || first_mem_access_qpi) == 1'b1) begin
							case(datain_reg[1:0])
								2'b00: dataout_reg <= dataout_reg_0;
								2'b01: dataout_reg <= dataout_reg_1;
								2'b10: dataout_reg <= dataout_reg_2;
								2'b11: dataout_reg <= dataout_reg_3;
							endcase 
						end
						else if ((mem_read || mem_read_dpi || mem_read_qpi) && addr3_phase == 0) begin
							dataout_reg <= mem_data[(MEM_ADD + addr_count) & addr_mask];
						end
					end
				default : begin

				end
			endcase
		end
	end	

	always @(negedge CLK or posedge CSn or posedge curr_mode)
	begin
		if(CSn == 1 ) begin
			hold_dq3_reg <=  1'b0;
			wp_dq2_reg <= 1'b0;
			so_dq1_reg <= 1'b0;
			si_dq0_reg <= 1'b0;
			dataout_bit_count <= 3'h0;
			dataout_bit_count_dpi <= 2'h0;
			dataout_bit_count_qpi <= 1'h0;
			
		end
		else if((dataout_phase == 1 && latency_phase == 0) || ( latency_count == (CR2[3:0]-1) && latency_phase ==1 ) || (mode_phase == 1 && (datain_bit_count == 0 || datain_bit_count_dpi == 0 || datain_bit_count_qpi == 0)  && CR2[3:0] == 0)) begin
			if (mode == SPIEN) begin
			so_dq1_reg <= dataout_reg[3'b111 - dataout_bit_count];
			dataout_bit_count <= dataout_bit_count + 1;
			end
			if (mode == DPIEN) begin
			so_dq1_reg <= dataout_reg[3'b111 -(2*dataout_bit_count_dpi)];
			si_dq0_reg <= dataout_reg[3'b110 -(2*dataout_bit_count_dpi)];
			dataout_bit_count_dpi <= dataout_bit_count_dpi + 1;
			end
			if (mode == QPIEN) begin
	         	hold_dq3_reg <= dataout_reg[3'b111 -(4*dataout_bit_count_qpi)];
	         	wp_dq2_reg <= dataout_reg[3'b110 - (4*dataout_bit_count_qpi)];
	        	so_dq1_reg <= dataout_reg[3'b101 - (4*dataout_bit_count_qpi)];
		        si_dq0_reg <= dataout_reg[3'b100 - (4*dataout_bit_count_qpi)];
                      // dataout_reg= dataout_reg<<4;
			dataout_bit_count_qpi <= dataout_bit_count_qpi + 1;
			end
	
		end
		else if((next == DATAOUT && next_state == 1 && (cmd == RDSR || cmd == RDCR)) || (next == DATAOUT && ((first_mem_access || first_mem_access_dpi || first_mem_access_qpi) == 1) && (cmd == READ  || cmd == FAST_READ))) begin
			if( cmd == RDSR ) begin
				so_dq1_reg <= SR[7];
				si_dq0_reg <= SR[7];
				wp_dq2_reg <= SR[7];
				hold_dq3_reg <= SR[7];
				dataout_bit_count <= dataout_bit_count + 1;
				dataout_bit_count_dpi <= dataout_bit_count_dpi + 1;
				dataout_bit_count_qpi <= dataout_bit_count_qpi + 1;

			end
			else if ( cmd == RDCR ) begin
	                  	hold_dq3_reg <= CR1[7];
                                wp_dq2_reg <= CR1[7];
                               	so_dq1_reg <= CR1[7];
				si_dq0_reg <= CR1[7];
				dataout_bit_count <= dataout_bit_count + 1;
				dataout_bit_count_dpi <= dataout_bit_count_dpi + 1;
				dataout_bit_count_qpi <= dataout_bit_count_qpi + 1;


			end
			else if ( cmd == READ  || cmd == FAST_READ ) begin
			//	if (mode == SPIEN)
			//	begin 
				case(datain_reg[1:0])
					2'b00: so_dq1_reg <= dataout_reg_0[7];
					2'b01: so_dq1_reg <= dataout_reg_1[7];
					2'b10: so_dq1_reg <= dataout_reg_2[7];
					2'b11: so_dq1_reg <= dataout_reg_3[7];
				endcase
                               
				dataout_bit_count <= 1;
				dataout_bit_count_dpi <= 1;
				dataout_bit_count_qpi <= 1;				
			//check ----
			//	end
			//	else if (mode==DPIEN)
			//	begin
			//	case(datain_reg[1:0])
			//		2'b00: so_dq1_reg <= dataout_reg_0[7] && si_dq0_reg <= dataout_reg_0[6]; 
			//		2'b01: so_dq1_reg <= dataout_reg_1[7] && si_dq0_reg <= dataout_reg_1[6]; 
			//		2'b10: so_dq1_reg <= dataout_reg_2[7] && si_dq0_reg <= dataout_reg_2[6];
			//		2'b11: so_dq1_reg <= dataout_reg_3[7] && si_dq0_reg <= dataout_reg_3[6];
			//	endcase
			//	end
			//	else if (mode==QPIEN) 
			//	begin
			//	case(datain_reg[1:0])
			//		2'b00: so_dq1_reg <= dataout_reg_0[7] && si_dq0_reg <= dataout_reg_0[6] && wp_dq2_reg<= dataout_reg_0[5] && hold_dq3_reg <= dataout_reg_0[4]; 
			//		2'b01: so_dq1_reg <= dataout_reg_1[7] && si_dq0_reg <= dataout_reg_1[6] && wp_dq2_reg<= dataout_reg_0[5] && hold_dq3_reg <= dataout_reg_0[4]; 
			//		2'b10: so_dq1_reg <= dataout_reg_2[7] && si_dq0_reg <= dataout_reg_2[6] && wp_dq2_reg<= dataout_reg_0[5] && hold_dq3_reg <= dataout_reg_0[4];  
			//		2'b11: so_dq1_reg <= dataout_reg_3[7] && si_dq0_reg <= dataout_reg_3[6] && wp_dq2_reg<= dataout_reg_0[5] && hold_dq3_reg <= dataout_reg_0[4];  
			//	endcase
			//	end
 
			end

		end	
		
	end
	
endmodule 

